@@ -211,7 +211,7 @@ def _isWhiteListedNotExistingModule( module_name ):
         "select", "__main__", "_winreg", "_warnings", "_sre", "_functools", "_hashlib",
         "_collections", "_locale", "_codecs", "_weakref", "_struct", "_dummy_threading",
         "binascii", "datetime", "_ast", "xxsubtype", "_bytesio", "cmath", "_fileio",
-        "aetypes", "aepack", "MacOS", "cd", "cl", "gdbm", "gl", "GL",
+        "aetypes", "aepack", "MacOS", "cd", "cl", "gdbm", "gl", "GL", "aetools",
 
         # Python-Qt4 does these if missing python3 parts:
         "PyQt4.uic.port_v3.string_io", "PyQt4.uic.port_v3.load_plugin",
@@ -265,7 +265,7 @@ def _isWhiteListedNotExistingModule( module_name ):
         "__package__.module", "__mangled_mod",
 
         # test_distutils.py
-        "distutils.tests",
+        "distutils.tests", "distutils.mwerkscompiler",
 
         # test_emails.py
         "email.test.test_email", "email.test.test_email_renamed",
@@ -293,4 +293,7 @@ def _isWhiteListedNotExistingModule( module_name ):
 
         # test_zipimport_support.py
         "test_zipped_doctest", "zip_pkg",
+
+        # test/test_zipimport_support.py
+        "test.test_cmd_line_script"
     )
