@@ -47,6 +47,46 @@ from .nodes.ParameterSpec import ParameterSpec
 from .nodes.FutureSpec import FutureSpec
 
 from .nodes import Nodes
+from .nodes.VariableRefNode import (
+    CPythonExpressionVariableRef,
+    CPythonExpressionTempVariableRef
+)
+from .nodes.ConstantRefNode import CPythonExpressionConstantRef
+from .nodes.BuiltinReferenceNodes import CPythonExpressionBuiltinExceptionRef
+from .nodes.ExceptionNodes import CPythonStatementRaiseException
+from .nodes.ComparisonNode import CPythonExpressionComparison
+from .nodes.ExecEvalNodes import CPythonStatementExec
+
+from .nodes.FunctionNodes import (
+    CPythonStatementFunctionBuilder,
+    CPythonExpressionLambdaBuilder,
+    CPythonExpressionFunctionBody
+)
+
+from .nodes.ContainerMakingNodes import (
+    CPythonExpressionKeyValuePair,
+    CPythonExpressionMakeSequence,
+    CPythonExpressionMakeDict,
+    CPythonExpressionMakeSet
+)
+
+from .nodes.StatementNodes import (
+    CPythonStatementsSequenceLoopBody,
+    CPythonStatementExpressionOnly,
+    CPythonStatementsSequence
+)
+
+from .nodes.ImportNodes import (
+    CPythonExpressionImportModule,
+    CPythonExpressionImportName,
+    CPythonStatementImportStar,
+)
+
+from .nodes.OperatorNodes import (
+    CPythonExpressionOperationBinary,
+    CPythonExpressionOperationUnary,
+    CPythonExpressionOperationNOT
+)
 
 import ast, sys
 
@@ -69,13 +109,13 @@ def pushDelayedWork( delayed_work ):
 
 
 def _buildConstantReferenceNode( constant, source_ref ):
-    return Nodes.CPythonExpressionConstantRef(
+    return CPythonExpressionConstantRef(
         constant   = constant,
         source_ref = source_ref
     )
 
 def _buildVariableReferenceNode( variable_name, source_ref ):
-    return Nodes.CPythonExpressionVariableRef(
+    return CPythonExpressionVariableRef(
         variable_name = variable_name,
         source_ref    = source_ref
     )
@@ -95,7 +135,7 @@ def buildStatementsNode( provider, nodes, source_ref, allow_none = False ):
     if not statements and allow_none:
         return None
 
-    return Nodes.CPythonStatementsSequence(
+    return CPythonStatementsSequence(
         statements = statements,
         source_ref = source_ref
     )
@@ -103,7 +143,7 @@ def buildStatementsNode( provider, nodes, source_ref, allow_none = False ):
 def buildLoopBodyNode( provider, nodes, source_ref ):
     statements = buildNodeList( provider, nodes, source_ref )
 
-    return Nodes.CPythonStatementsSequenceLoopBody(
+    return CPythonStatementsSequenceLoopBody(
         statements = statements,
         source_ref = source_ref
     )
@@ -187,7 +227,7 @@ def buildFunctionNode( provider, node, source_ref ):
     decorators = buildDecoratorNodes( provider, node.decorator_list, source_ref )
     defaults = buildNodeList( provider, node.args.defaults, source_ref )
 
-    result = Nodes.CPythonStatementFunctionBuilder(
+    result = CPythonStatementFunctionBuilder(
         target     = buildVariableRefAssignTarget( node.name, source_ref ),
         defaults   = defaults,
         decorators = decorators,
@@ -200,7 +240,7 @@ def buildFunctionNode( provider, node, source_ref ):
         while real_provider.isExpressionClassBody():
             real_provider = real_provider.provider
 
-        function_body = Nodes.CPythonExpressionFunctionBody(
+        function_body = CPythonExpressionFunctionBody(
             provider   = real_provider,
             name       = node.name,
             doc        = function_doc,
@@ -228,7 +268,7 @@ def isSameListContent( a, b ):
 def buildLambdaNode( provider, node, source_ref ):
     defaults = buildNodeList( provider, node.args.defaults, source_ref )
 
-    result = Nodes.CPythonExpressionLambdaBuilder(
+    result = CPythonExpressionLambdaBuilder(
         defaults   = defaults,
         source_ref = source_ref,
     )
@@ -239,7 +279,7 @@ def buildLambdaNode( provider, node, source_ref ):
         while real_provider.isExpressionClassBody():
             real_provider = real_provider.provider
 
-        function_body = Nodes.CPythonExpressionFunctionBody(
+        function_body = CPythonExpressionFunctionBody(
             provider   = real_provider,
             name       = "lambda",
             doc        = None,
@@ -256,7 +296,7 @@ def buildLambdaNode( provider, node, source_ref ):
         )
 
         if function_body.isGenerator():
-            body = Nodes.CPythonStatementExpressionOnly(
+            body = CPythonStatementExpressionOnly(
                 expression = Nodes.CPythonExpressionYield(
                     expression = body,
                     for_return = True,
@@ -270,7 +310,7 @@ def buildLambdaNode( provider, node, source_ref ):
                 source_ref = function_body.getSourceReference()
             )
 
-        body = Nodes.CPythonStatementsSequence(
+        body = CPythonStatementsSequence(
             statements = ( body, ),
             source_ref = function_body.getSourceReference()
         )
@@ -312,9 +352,10 @@ def buildWhileLoopNode( provider, node, source_ref ):
 def buildFunctionCallNode( provider, node, source_ref ):
     positional_args = buildNodeList( provider, node.args, source_ref )
 
-    # TODO: Clarify if the source_ref could be better
+    # Only the values of keyword pairs have a real source ref, and those only really
+    # matter, so that makes sense.
     pairs = [
-        Nodes.CPythonExpressionKeyValuePair(
+        CPythonExpressionKeyValuePair(
             key        = _buildConstantReferenceNode(
                 constant   = keyword.arg,
                 source_ref = source_ref
@@ -356,7 +397,7 @@ def buildSequenceCreationNode( provider, node, source_ref ):
             source_ref = source_ref
         )
     else:
-        return Nodes.CPythonExpressionMakeSequence(
+        return CPythonExpressionMakeSequence(
             sequence_kind = sequence_kind,
             elements      = elements,
             source_ref    = source_ref
@@ -402,9 +443,9 @@ def buildDictionaryNode( provider, node, source_ref ):
             source_ref = source_ref
         )
     else:
-        return Nodes.CPythonExpressionMakeDict(
+        return CPythonExpressionMakeDict(
             pairs      = [
-                Nodes.CPythonExpressionKeyValuePair( key, value, key.getSourceReference() )
+                CPythonExpressionKeyValuePair( key, value, key.getSourceReference() )
                 for key, value in
                 zip( keys, values )
             ],
@@ -430,7 +471,7 @@ def buildSetNode( provider, node, source_ref ):
             source_ref = source_ref
         )
     else:
-        return Nodes.CPythonExpressionMakeSet(
+        return CPythonExpressionMakeSet(
             values     = values,
             source_ref = source_ref
         )
@@ -702,7 +743,7 @@ def _buildContractionNode( provider, node, builder_class, body_class, list_contr
             )
 
             contraction_body.setBody(
-                Nodes.CPythonExpressionKeyValuePair(
+                CPythonExpressionKeyValuePair(
                     key        = key_node,
                     value      = value_node,
                     source_ref = source_ref
@@ -727,7 +768,7 @@ def _buildContractionNode( provider, node, builder_class, body_class, list_contr
         # it to a lambda generator function, something that does not exist in CPython, but
         # for which we can generator code.
         if contraction_body.isExpressionGeneratorBody() and contraction_body.isGenerator():
-            generator_function_body = Nodes.CPythonExpressionFunctionBody(
+            generator_function_body = CPythonExpressionFunctionBody(
                 provider   = provider,
                 name       = "pseudo",
                 doc        = None,
@@ -735,7 +776,7 @@ def _buildContractionNode( provider, node, builder_class, body_class, list_contr
                 source_ref = source_ref
             )
 
-            body = Nodes.CPythonStatementExpressionOnly(
+            body = CPythonStatementExpressionOnly(
                 expression = Nodes.CPythonExpressionYield(
                     expression = contraction_body.getBody(),
                     for_return = False,
@@ -744,12 +785,17 @@ def _buildContractionNode( provider, node, builder_class, body_class, list_contr
                 source_ref = source_ref
             )
 
-            body = Nodes.CPythonStatementsSequence(
+            body = CPythonStatementsSequence(
                 statements = ( body, ),
                 source_ref = source_ref
             )
 
-            sources = [ Nodes.CPythonExpressionVariableRef( variable_name = "_iterated", source_ref = source_ref ) ]
+            sources = [
+                CPythonExpressionVariableRef(
+                    variable_name = "_iterated",
+                    source_ref    = source_ref
+                )
+            ]
             sources += contraction_body.getSources()
 
             for target, source, condition in zip(
@@ -764,7 +810,7 @@ def _buildContractionNode( provider, node, builder_class, body_class, list_contr
                     source_ref = source_ref
                 )
 
-                body = Nodes.CPythonStatementsSequenceLoopBody(
+                body = CPythonStatementsSequenceLoopBody(
                     statements = ( body, ),
                     source_ref = source_ref
                 )
@@ -777,7 +823,7 @@ def _buildContractionNode( provider, node, builder_class, body_class, list_contr
                     source_ref = source_ref,
                 )
 
-                body = Nodes.CPythonStatementsSequence(
+                body = CPythonStatementsSequence(
                     statements = ( body, ),
                     source_ref = source_ref
                 )
@@ -785,7 +831,7 @@ def _buildContractionNode( provider, node, builder_class, body_class, list_contr
             generator_function_body.setBody( body )
             generator_function_body.markAsGenerator()
 
-            new_result = Nodes.CPythonExpressionLambdaBuilder(
+            new_result = CPythonExpressionLambdaBuilder(
                 defaults   = (),
                 source_ref = source_ref,
             )
@@ -850,18 +896,70 @@ def buildGeneratorExpressionNode( provider, node, source_ref ):
     )
 
 def buildComparisonNode( provider, node, source_ref ):
-    comparison = [ buildNode( provider, node.left, source_ref ) ]
-
     assert len( node.comparators ) == len( node.ops )
 
-    for comparator, operand in zip( node.ops, node.comparators ):
-        comparison.append( getKind( comparator ) )
-        comparison.append( buildNode( provider, operand, source_ref ) )
+    # The operands are split out
+    left = buildNode( provider, node.left, source_ref )
+    rights = [
+        buildNode( provider, comparator, source_ref )
+        for comparator in
+        node.comparators
+    ]
 
-    return Nodes.CPythonExpressionComparison(
-        comparison = comparison,
-        source_ref = source_ref
-    )
+    # Only the first comparison has as left operands as the real thing, the others must
+    # reference the previous comparison right one temp variable ref.
+    result = []
+
+    # For PyLint to like it, this will hold the previous one, normally.
+    tmp_variable = None
+
+    for comparator, right in zip( node.ops, rights ):
+        if result:
+            # Now we know it's not the only one, so we change the "left" to be a reference
+            # to the previously saved right side.
+            left = CPythonExpressionTempVariableRef(
+                variable   = tmp_variable.makeReference( provider ),
+                source_ref = source_ref
+            )
+
+            tmp_variable = None
+
+        if right is not rights[-1]:
+            # Now we known it's not the last one, so we ought to preseve the "right" so it
+            # can be referenced by the next part that will come. We do it by assining it
+            # to a temp variable to be shared with the next part.
+            tmp_variable = provider.getTempVariable()
+
+            right = Nodes.CPythonExpressionAssignment(
+                source     = right,
+                target     = Nodes.CPythonAssignTargetVariable(
+                    variable_ref = CPythonExpressionTempVariableRef(
+                        variable   = tmp_variable.makeReference( provider ),
+                        source_ref = source_ref
+                    ),
+                    source_ref   = source_ref
+                ),
+                source_ref = source_ref
+            )
+
+        result.append(
+            CPythonExpressionComparison(
+                left       = left,
+                right      = right,
+                comparator = getKind( comparator ),
+                source_ref = source_ref
+            )
+        )
+
+    assert tmp_variable is None
+
+    if len( result ) > 1:
+        return Nodes.CPythonExpressionBoolAND(
+            operands   = result,
+            source_ref = source_ref
+        )
+    else:
+        return result[ 0 ]
 
 def buildConditionNode( provider, node, source_ref ):
     return Nodes.CPythonStatementConditional(
@@ -910,7 +1008,7 @@ _has_raise_value = Utils.getPythonVersion() < 300
 
 def buildRaiseNode( provider, node, source_ref ):
     if _has_raise_value:
-        return Nodes.CPythonStatementRaiseException(
+        return CPythonStatementRaiseException(
             exception_type  = buildNode( provider, node.type, source_ref, True ),
             exception_value = buildNode( provider, node.inst, source_ref, True ),
             exception_trace = buildNode( provider, node.tback, source_ref, True ),
@@ -919,7 +1017,7 @@ def buildRaiseNode( provider, node, source_ref ):
     else:
         assert node.cause is None
 
-        return Nodes.CPythonStatementRaiseException(
+        return CPythonStatementRaiseException(
             exception_type  = buildNode( provider, node.exc, source_ref, True ),
             exception_value = None,
             exception_trace = None,
@@ -937,14 +1035,14 @@ def buildAssertNode( provider, node, source_ref ):
     # Therefore assert statements are really just conditional statements with a static
     # raise contained.
     return Nodes.CPythonStatementConditional(
-        condition = Nodes.CPythonExpressionOperationNOT(
+        condition = CPythonExpressionOperationNOT(
             operand    = buildNode( provider, node.test, source_ref ),
             source_ref = source_ref
         ),
-        yes_branch = Nodes.CPythonStatementsSequence(
+        yes_branch = CPythonStatementsSequence(
             statements = (
-                Nodes.CPythonStatementRaiseException(
-                    exception_type = Nodes.CPythonExpressionBuiltinExceptionRef(
+                CPythonStatementRaiseException(
+                    exception_type = CPythonExpressionBuiltinExceptionRef(
                         exception_name = "AssertionError",
                         source_ref     = source_ref
                     ),
@@ -992,7 +1090,7 @@ def _buildExtSliceNode( provider, node, source_ref ):
 
         elements.append( element )
 
-    return Nodes.CPythonExpressionMakeSequence(
+    return CPythonExpressionMakeSequence(
         sequence_kind = "TUPLE",
         elements      = elements,
         source_ref    = source_ref
@@ -1076,7 +1174,7 @@ def _buildImportModulesNode( import_names, source_ref ):
         )
 
         if local_name:
-            import_node = Nodes.CPythonExpressionImportModule(
+            import_node = CPythonExpressionImportModule(
                 module_name = module_name,
                 import_list = None,
                 level       = -1, # TODO: Correct?!
@@ -1084,13 +1182,13 @@ def _buildImportModulesNode( import_names, source_ref ):
             )
 
             for import_name in module_name.split(".")[1:]:
-                import_node = Nodes.CPythonExpressionImportName(
+                import_node = CPythonExpressionImportName(
                     module      = import_node,
                     import_name = import_name,
                     source_ref  = source_ref
                 )
         else:
-            import_node = Nodes.CPythonExpressionImportModule(
+            import_node = CPythonExpressionImportModule(
                 module_name = module_name,
                 import_list = None,
                 level       = -1, # TODO: Correct?!
@@ -1108,7 +1206,7 @@ def _buildImportModulesNode( import_names, source_ref ):
     # Note: Each import is sequential. It can succeed, and the failure of a later one is
     # not changing one. We can therefore have a sequence of imports that only import one
     # thing therefore.
-    return Nodes.CPythonStatementsSequence(
+    return CPythonStatementsSequence(
         statements = import_nodes,
         source_ref = source_ref
     )
@@ -1171,8 +1269,8 @@ def buildImportFromNode( provider, node, source_ref ):
         imports.append( object_name )
 
     if None in targets:
-        return Nodes.CPythonStatementImportStar(
-            module_import = Nodes.CPythonExpressionImportModule(
+        return CPythonStatementImportStar(
+            module_import = CPythonExpressionImportModule(
                 module_name = module_name,
                 import_list = ( "*", ),
                 level       = level,
@@ -1187,8 +1285,8 @@ def buildImportFromNode( provider, node, source_ref ):
             import_nodes.append(
                 Nodes.CPythonStatementAssignment(
                     targets    = ( target, ),
-                    expression = Nodes.CPythonExpressionImportName(
-                        module      = Nodes.CPythonExpressionImportModule(
+                    expression = CPythonExpressionImportName(
+                        module      = CPythonExpressionImportModule(
                             module_name = module_name,
                             import_list = imports,
                             level       = level,
@@ -1204,7 +1302,7 @@ def buildImportFromNode( provider, node, source_ref ):
         # Note: Each import is sequential. It can succeed, and the failure of a later one is
         # not changing one. We can therefore have a sequence of imports that only import one
         # thing therefore.
-        return Nodes.CPythonStatementsSequence(
+        return CPythonStatementsSequence(
             statements = import_nodes,
             source_ref = source_ref
         )
@@ -1244,10 +1342,7 @@ def buildExecNode( provider, node, source_ref ):
         if globals_node.isExpressionConstantRef() and globals_node.getConstant() is None:
             globals_node = None
 
-    if provider.isExpressionFunctionBody():
-        provider.markAsExecContaining()
-
-    return Nodes.CPythonStatementExec(
+    return CPythonStatementExec(
         source_code = buildNode( provider, body, source_ref ),
         globals_arg = globals_node,
         locals_arg  = locals_node,
@@ -1324,7 +1419,7 @@ def buildBoolOpNode( provider, node, source_ref ):
         )
     elif bool_op == "Not":
         # The "not" is really only a unary operation and no special.
-        return Nodes.CPythonExpressionOperationNOT(
+        return CPythonExpressionOperationNOT(
             operand    = buildNode( provider, node.operand, source_ref ),
             source_ref = source_ref
         )
@@ -1378,7 +1473,7 @@ def buildYieldNode( provider, node, source_ref ):
         )
 
 def buildExprOnlyNode( provider, node, source_ref ):
-    return Nodes.CPythonStatementExpressionOnly(
+    return CPythonStatementExpressionOnly(
         expression = buildNode( provider, node.value, source_ref ),
         source_ref = source_ref
     )
@@ -1392,7 +1487,7 @@ def buildUnaryOpNode( provider, node, source_ref ):
             source_ref = source_ref
         )
     else:
-        return Nodes.CPythonExpressionOperationUnary(
+        return CPythonExpressionOperationUnary(
             operator   = getKind( node.op ),
             operand    = buildNode( provider, node.operand, source_ref ),
             source_ref = source_ref
@@ -1405,7 +1500,7 @@ def buildBinaryOpNode( provider, node, source_ref ):
     if operator == "Div" and source_ref.getFutureSpec().isFutureDivision():
         operator = "TrueDiv"
 
-    return Nodes.CPythonExpressionOperationBinary(
+    return CPythonExpressionOperationBinary(
         operator   = operator,
         left       = buildNode( provider, node.left, source_ref ),
         right      = buildNode( provider, node.right, source_ref ),
@@ -1413,7 +1508,7 @@ def buildBinaryOpNode( provider, node, source_ref ):
     )
 
 def buildReprNode( provider, node, source_ref ):
-    return Nodes.CPythonExpressionOperationUnary(
+    return CPythonExpressionOperationUnary(
         operator   = "Repr",
         operand    = buildNode( provider, node.value, source_ref ),
         source_ref = source_ref
@@ -1599,18 +1694,21 @@ def buildModuleTree( filename, package, is_main ):
         if is_main:
             module_name = "__main__"
         else:
-            module_name = Utils.basename( filename ).replace( ".py", "" )
+            module_name = Utils.basename( filename )
+
+            if module_name.endswith( ".py" ):
+                module_name = module_name[:-3]
 
         result = Nodes.CPythonModule(
             name       = module_name,
             package    = package,
             source_ref = source_ref
         )
-    elif Utils.isDir( filename ) and Utils.isFile( filename + "/__init__.py" ):
-        source_filename = filename + "/__init__.py"
+    elif Utils.isDir( filename ) and Utils.joinpath( filename, "__init__.py" ):
+        source_filename = Utils.joinpath( filename, "__init__.py" )
 
         source_ref = SourceCodeReferences.fromFilename(
-            filename    = Utils.abspath( filename ) + "/__init__.py",
+            filename    = Utils.abspath( source_filename ),
             future_spec = FutureSpec()
         )
 
@@ -1623,11 +1721,12 @@ def buildModuleTree( filename, package, is_main ):
         sys.stderr.write(  "Nuitka: can't open file '%s'.\n" % filename )
         sys.exit( 2 )
 
-    buildParseTree(
-        provider    = result,
-        source_code = open( source_filename, "rU" ).read(),
-        source_ref  = source_ref,
-        replacement = False,
-    )
+    with open( source_filename, "rU" ) as source_file:
+        buildParseTree(
+            provider    = result,
+            source_code = source_file.read(),
+            source_ref  = source_ref,
+            replacement = False,
+        )
 
     return result
